from heapq import heapify, heappush, heappop
from vbweights import w
from comptab import fcomp
from bitcoding import B_dict
from inverse import inv
import itertools
from counters import counter
from assignment import add_var, remove_var, mrv_variables
from itertools import combinations
from learned_weights import updateWeights
from glob import globs

if globs['split'] == 0: # splitting based on set of base relations
   from basesplit import bsplit as split
elif globs['split'] == 1: # splitting based on horn set
   from hornsplit import hsplit as split

DALL = B_dict['DALL']
cardmult = max(w)

# path consistency using van Beek weights
def PCw(ConMatrix, neighbors, m, n, repairs, trail):
   pq = []                         # list of entries arranged in a heap
   entry_finder = {}               # mapping of tasks to entries
   #REMOVED = None                  # placeholder for a removed task
   pcounter = itertools.count()     # unique sequence count
   Vars = len(ConMatrix)

   def add_task(task, priority=0):
       'Add a new task or update the priority of an existing task'
       if task in entry_finder:
           remove_task(task)
       count = next(pcounter)
       entry = [priority, count, task]
       entry_finder[task] = entry
       heappush(pq, entry)

   def remove_task(task):
       'Mark an existing task as REMOVED.  Raise KeyError if not found.'
       entry = entry_finder.pop(task)
       entry[-1] = None

   def pop_task():
       'Remove and return the lowest priority task. Raise KeyError if empty.'
       while pq:
           task = heappop(pq)[-1]
           if task:
               del entry_finder[task]
               return task
       raise KeyError('pop from an empty priority queue')   

   add_task((m,n),w[ConMatrix[m][n]-1]) 

   # as long as the queue is not empty, process it
   while pq:

      try:
         (i,j) = pop_task() # grab the appropriate relation
      except KeyError:
         break
      
      #counter['arcs'] += 1 # increment visited arcs counter 
  
      # create all triplets to be checked for path consistency
      for k in neighbors[i] & neighbors[j]:

          temp = fcomp[ConMatrix[k][i]-1][ConMatrix[i][j]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
  
             # constrain arc (k,i,j)
             temp = temp & ConMatrix[k][j]
             if temp != ConMatrix[k][j]:
                if not temp: 
                   updateWeights(list(combinations([i,j,k],2)),counter['decisions'])
                   return False # inconsistency
                r = ConMatrix[k][j]
                ConMatrix[k][j] = temp
                ConMatrix[j][k] = inv[temp-1]
                if k < j:
                   add_task((k, j), cardmult*split[ConMatrix[k][j]-1][0] + w[ConMatrix[k][j]-1])   
                   if (k, j) not in repairs:
                      repairs[(k, j)] = r
                   if (k,j) not in mrv_variables:
                       if split[ConMatrix[k][j]-1][0] > 1:
                         add_var((k,j))
                         trail.append((1,(k,j)))
                   else:
                      if split[ConMatrix[k][j]-1][0] <= 1:
                         remove_var((k,j))
                         trail.append((0,(k,j)))
                else:
                   add_task((j, k), cardmult*split[ConMatrix[j][k]-1][0] + w[ConMatrix[j][k]-1])
                   if (j, k) not in repairs:
                      repairs[(j, k)] = inv[r-1]
                   if (j,k) not in mrv_variables:
                      if split[ConMatrix[j][k]-1][0] > 1:
                         add_var((j,k))
                         trail.append((1,(j,k)))
                   else:
                      if split[ConMatrix[j][k]-1][0] <= 1:
                         remove_var((j,k))
                         trail.append((0,(j,k)))

          temp = fcomp[ConMatrix[i][j]-1][ConMatrix[j][k]-1]
          if temp != DALL:

             #counter['cons'] += 1 # increment checked constraints counter
          
             # constrain arc (i,j,k) 
             temp = temp & ConMatrix[i][k] 
             if temp != ConMatrix[i][k]:
                if not temp: 
                   updateWeights(list(combinations([i,j,k],2)),counter['decisions'])
                   return False # inconsistency
                r = ConMatrix[i][k]
                ConMatrix[i][k] = temp
                ConMatrix[k][i] = inv[temp-1]
                if i < k:
                   add_task((i, k), cardmult*split[ConMatrix[i][k]-1][0] + w[ConMatrix[i][k]-1])
                   if (i, k) not in repairs:
                      repairs[(i, k)] = r
                   if (i,k) not in mrv_variables:
                      if split[ConMatrix[i][k]-1][0] > 1: 
                         add_var((i,k))
                         trail.append((1,(i,k)))
                   else:
                      if split[ConMatrix[i][k]-1][0] <= 1:
                         remove_var((i,k))
                         trail.append((0,(i,k)))
                else:
                   add_task((k, i), cardmult*split[ConMatrix[k][i]-1][0] + w[ConMatrix[k][i]-1])
                   if (k, i) not in repairs:
                      repairs[(k, i)] = inv[r-1]
                   if (k,i) not in mrv_variables:
                      if split[ConMatrix[k][i]-1][0] > 1:
                         add_var((k,i))
                         trail.append((1,(k,i)))
                   else:
                      if split[ConMatrix[k][i]-1][0] <= 1:
                         remove_var((k,i))
                         trail.append((0,(k,i))) 

   # the network is concistent and can't be refined further
   return True    
